/* Copyright 2009 privacychoice LLC */

const IOptOut = Components.interfaces.IOptOut;

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

const MODE_RDONLY   = 0x01;
const PERMS_FILE    = 0644;

const cookieManager = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager);
const cookieManager2 = Cc["@mozilla.org/cookiemanager;1"].getService(Ci.nsICookieManager2);
const ioService = Cc['@mozilla.org/network/io-service;1'].getService(Ci.nsIIOService);
const observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
const prefBranch = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService)
                                                           .getBranch("extensions.privacychoice.optout.");

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

function OptOut() {
}

OptOut.prototype = {
  pcDomain: "www.privacychoice.org",
  adNetworks: null,
  optoutList: null,
  cookieName: "uid",
  removingAddonCookie: false,
  beingUninstalled: false,
  
  log: function(string) {
    var aConsoleService = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);   
    aConsoleService.logStringMessage(string);
  },

  /* write out optout cookies */
  updateOptOutCookies: function() {
    if (this.adNetworks && this.optoutList) {
	  var domains = [];
      for (var i in this.adNetworks) {
        if (this.optoutList.match("," + i + ",")) {
		  this.writeAdNetworkCookie(i);
        }
      }
    }
  },

  /* loads adnetwork list from file and puts it in an array */
  loadAdNetworkList: function() {
    /* This is normally how the network list would be initialized */
    file = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties)
                                                      .get("ProfD", Ci.nsILocalFile);
    file.append("privacychoice");
    file.append("privacy.json");
    if (!file.exists()) {
      this.getAdNetworkList();
      return;
    }
    
    var fileInStream = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
    var cis = Cc["@mozilla.org/intl/converter-input-stream;1"].createInstance(Ci.nsIConverterInputStream);

    fileInStream.init(file, MODE_RDONLY, PERMS_FILE, false);
    cis.init(fileInStream,  null, file.fileSize, Ci.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
    var jsonFile = {value:null};
    cis.readString(file.fileSize, jsonFile);
    cis.close();

    var jsonAdNetworks = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON)
                                                      .decode(jsonFile.value);
    this.adNetworks = [];
    for (var i=0; i < jsonAdNetworks.length; i++) {
      if (!this.adNetworks[jsonAdNetworks[i].ad_network_id]) {
        this.adNetworks[jsonAdNetworks[i].ad_network_id] = [];
      }
      this.adNetworks[jsonAdNetworks[i].ad_network_id].push(jsonAdNetworks[i]);
    }
  },

  /* Retrieves adnetwork list from server */
  getAdNetworkList: function(timestamp) {
    var outputFile = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties)
                                                                .get("ProfD", Ci.nsILocalFile);

    outputFile.append("privacychoice");
    if (!outputFile.exists()) {
      outputFile.create(Ci.nsIFile.DIRECTORY_TYPE, 0755);
    }

    outputFile.append("privacy.json");

    var uri = ioService.newURI("http://" + this.pcDomain + "/ad_networks/export_json", null, null);
    
    var channel = ioService.newChannelFromURI(uri);
    var downloader = Cc["@mozilla.org/network/downloader;1"].createInstance(Ci.nsIDownloader);
    var _this = this;
    var listener = {
      onDownloadComplete: function(downloader, request, ctxt, status, result) {
        if (Components.isSuccessCode(status)) {
          /* set timestamp here since download might have failes */
          prefBranch.setIntPref("timestamp", timestamp);
          result.QueryInterface(Ci.nsIFile);
        }
        /* Might figure out a way later to pass result */
        _this.loadAdNetworkList();
        _this.updateOptOutCookies();
      }
    }
    downloader.init(listener, outputFile);
    channel.asyncOpen(downloader, null);
  },

  /* XHR to server to get current JSON file based on browser identifier */
  getCurrentOptOutInfo: function() {
    var browserID = this.getBrowserID();
    if (!browserID) {
      return;
    }

    var request = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Ci.nsIXMLHttpRequest)
  
    request.open("GET", "http://" + this.pcDomain + "/privacy/get_optout_list/" + browserID);

    var _this = this;

    request.onreadystatechange = function() {
      if (request.readyState == 4) {
        if (request.status == 200) {
          var optoutInfo  = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON)
                                                         .decode(request.responseText);
		  if (optoutInfo.optout_list) {
            _this.optoutList = "," + optoutInfo.optout_list + ",";
            prefBranch.setCharPref("optoutList", optoutInfo.optout_list);
		  } else {
            _this.optoutList = ",,"; // For consistency with other code
            prefBranch.setCharPref("optoutList", "");			
		  }
        } else {
          try {
            _this.optoutList = "," + prefBranch.getCharPref("optoutList") + ",";
          } catch (ex) {
          }
        }
        var timestamp = prefBranch.getIntPref("timestamp");
        if (optoutInfo && (timestamp != optoutInfo.networks_last_update)) {
          _this.getAdNetworkList(optoutInfo.networks_last_update);
        } else {
          if (!_this.adNetworks) {
            _this.loadAdNetworkList();
          }
          _this.updateOptOutCookies();
        }
      }
    }
    request.send(null);
  },

  removeAllAdNetworkCookies: function() {
    for (var i in this.adNetworks) {
      this.removeAdNetworkCookie(i, false);
    }
  },

  removeAdNetworkCookie: function(adNetworkID) {
    for (let i=0; i < this.adNetworks[adNetworkID].length; i++) {
      try {
        cookieManager.remove(this.adNetworks[adNetworkID][i].exact_domain,
                             this.adNetworks[adNetworkID][i].keyname,
                             "/",
                             false);
      } catch (ex) {}
      try {
        cookieManager.remove("." + this.adNetworks[adNetworkID][i].exact_domain,
                             this.adNetworks[adNetworkID][i].keyname,
                             "/",
                             false);
      } catch (ex) {}
    }
  },

  writeAdNetworkCookie: function(adNetworkID) {
    for (let i=0; i < this.adNetworks[adNetworkID].length; i++) {
	  if (this.adNetworks[adNetworkID][i].exact_domain) {
        this.writeCookie(this.adNetworks[adNetworkID][i].keyname,
                         this.adNetworks[adNetworkID][i].data,
                         this.adNetworks[adNetworkID][i].exact_domain);
	  }
	  if (this.adNetworks[adNetworkID][i].lso_domain) {
        this.removeFlashCookies(this.adNetworks[adNetworkID][i].lso_domain);
	  }
    }
  },

  /* generic function for writing cookies */
  writeCookie: function(name, value, domain) {
    try {
      var now = new Date();
      var expired = (new Date(Date.parse("January 1, " + (now.getFullYear() + 10) + " 12:00:00"))).getTime() / 1000;
	  cookieManager2.add(domain, "/", name, value, false, false, false, expired);
    } catch (ex) {
	  alert(ex);
    }
  },

  /* If we've lost the privacychoice cookie, rewrite it from prefs */
  rewriteOptOutCookie: function() {
    var browserID;
    try {
      browserID = prefBranch.getCharPref("browserID");
      this.writeCookie(this.cookieName, browserID, this.pcDomain);
    } catch (ex) {
//      Components.utils.reportError("rewriteOptOutCookie: no browserID preference");
    }
    return browserID;
  },

  getBrowserID: function() {
    var browserID;

	var e = cookieManager.enumerator;
	while (e.hasMoreElements()) {
	  var cookie = e.getNext();
	  if (cookie && cookie instanceof Components.interfaces.nsICookie) {
		if ((cookie.host == this.pcDomain) || (cookie.host == "." + this.pcDomain)) {
		  if (cookie.name == this.cookieName) {
			browserID = cookie.value;
			break;
		  }
		}
	  }
	}

    if (!browserID) {
      browserID = this.rewriteOptOutCookie();
    } else {
      prefBranch.setCharPref("browserID", browserID);
    }
    return browserID;
  },

  /* Remove
        FLASHDIR/#SharedObject/XXXXXX/domainname
        FLASHDIR/macromedia.com/support/flashplayer/sys/#domainname
        FLASHDIR/macromedia.com/support/flashplayer/sys/##/domainname
 */

  removeFlashCookies: function(domain) {
	if (domain.charAt(0) == '.') {
	  domain = domain.substr(1);
	  var sysDir = this.getFlashDir();
	  sysDir.append("macromedia.com")
	  sysDir.append("support")
	  sysDir.append("flashplayer")
	  sysDir.append("sys")
	  if (sysDir.exists()) {
		var e = sysDir.directoryEntries;
		while (e.hasMoreElements()) {
		  var f = e.getNext().QueryInterface(Components.interfaces.nsIFile);
		  if (f.leafName.match(domain)) {
			f.remove(true);
		  }
		}
		sysDir.append("##")
		if (sysDir.exists()) {
		  var e = sysDir.directoryEntries;
		  while (e.hasMoreElements()) {
			var f = e.getNext().QueryInterface(Components.interfaces.nsIFile);
			if (f.leafName.match(domain)) {
			  f.remove(true);
			}
		  }
		}
	  }
	  var sharedObjDir = this.getFlashDir();
	  sharedObjDir.append("#SharedObjects");
	  if (sharedObjDir.exists()) {
		e = sharedObjDir.directoryEntries;
		while (e.hasMoreElements()) {
		  var f = e.getNext().QueryInterface(Components.interfaces.nsIFile);
		  if (f.isDirectory()) {
			var de = f.directoryEntries;
			while (de.hasMoreElements()) {
			  var f2 = de.getNext().QueryInterface(Components.interfaces.nsIFile);
			  if (f2.leafName.match(domain)) {
				f2.remove(true);
			  }
			}
		  }
		}
	  }
	} else {
	  var sysDir = this.getFlashDir();
	  sysDir.append("macromedia.com");
	  sysDir.append("support");
	  sysDir.append("flashplayer");
	  sysDir.append("sys");
	  sysDir.append("#" + domain);
	  if (sysDir.exists()) {
		sysDir.remove(true);
	  }
	  sysDir.leafName = "##";
	  sysDir.append(domain);
	  if (sysDir.exists()) {
		sysDir.remove(true);
	  }
	  var sharedObjDir = this.getFlashDir();
	  sharedObjDir.append("#SharedObjects");
	  if (sharedObjDir.exists()) {
		var e = sharedObjDir.directoryEntries;
		while (e.hasMoreElements()) {
		  var f = e.getNext().QueryInterface(Components.interfaces.nsIFile);
		  if (f.isDirectory()) {
			f.append(domain);
			if (f.exists()) {
			  f.remove(true);
			}
		  }
		}
	  }
	}
  },
  
  getFlashDir: function() {
	var flashdir;
	try {
		flashdir = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties)
														      .get("UsrPrfs", Components.interfaces.nsILocalFile);
		flashdir.append("Macromedia");
		flashdir.append("Flash Player");
	} catch (ex) {}
	if (!flashdir) {
		try {
			flashdir = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties)
															      .get("AppData", Components.interfaces.nsILocalFile);
		flashdir.append("Macromedia");
		flashdir.append("Flash Player");            
		} catch (ex) {}
	}
	if (!flashdir) {
		try {
			flashdir = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties)
															      .get("Home", Components.interfaces.nsILocalFile);
		flashdir.append(".macromedia");
		flashdir.append("Flash_Player");            
		} catch (ex) {}
	}
	return flashdir;
  },

  addAddonCookie: function() {
	try {
      var now = new Date();
      var expired = (new Date(Date.parse("January 1, " + (now.getFullYear() + 10) + " 12:00:00"))).getTime() / 1000;
	  cookieManager2.add(this.pcDomain, "/", "addon", "true", false, false, true, expired);
	} catch (ex) {}
  },

  removeAddonCookie: function() {
	this.removingAddonCookie = true;
	try {
	  cookieManager.remove(this.pcDomain,
						   "addon",
						   "/",
						   false);
	} catch (ex) {
	}
	try {
	  cookieManager.remove("." + this.pcDomain,
						   "addon",
						   "/",
						   false);
	} catch (ex) {
	}
	this.removingAddonCookie = false;
  },

  observe: function(subject, topic, data) {
    switch (topic) {
      case "cookie-changed":
        if (data == 'cleared') {
		  this.addAddonCookie();
          // All cookies have been removed
          // Rewrite browser identifier cookie
          this.rewriteOptOutCookie();
          this.getCurrentOptOutInfo();
        } else if (data == 'deleted') {
          // One cookie has been removed
          // See if it is the privacy choice cookie
          var cookie = subject.QueryInterface(Ci.nsICookie);
          if ((cookie.host == this.pcDomain) || (cookie.host == "." + this.pcDomain)) {
            if (cookie.name == this.cookieName) {
              this.rewriteOptOutCookie();
            } else if ((cookie.name == "addon") && (!this.removingAddonCookie)) {
			  this.addAddonCookie();
            }
          }
        }
        break;

      case "app-startup":
        observerService.addObserver(this, "profile-after-change", false);
        observerService.addObserver(this, "profile-before-change", false);
        observerService.addObserver(this, "cookie-changed", false);
        observerService.addObserver(this,"xpcom-shutdown",false);
        observerService.addObserver(this, "em-action-requested", false);
        break;

      case "profile-after-change":
        try {
          this.pcDomain = prefBranch.getCharPref("server");
        } catch (ex) {}
		this.addAddonCookie();
        var browserID = this.getBrowserID();
        if (browserID) {
          this.getCurrentOptOutInfo();
        } else {
          prefBranch.setBoolPref("firstrun", true);
        }
        break;

      case "profile-before-change":
        this.removeAddonCookie();
		if (this.beingUninstalled) {
		  prefBranch.clearUserPref("installedVersion");
		}
        break;

      case "em-action-requested":
        subject.QueryInterface(Components.interfaces.nsIUpdateItem);
		if (subject.id == "addon@privacychoice.org") {
		  switch (data) {
			case "item-uninstalled":
			  this.beingUninstalled = true;
			  var gBrowser = Cc['@mozilla.org/appshell/window-mediator;1'].getService(Ci.nsIWindowMediator)
																		  .getMostRecentWindow('navigator:browser')
																		  .gBrowser;
			  gBrowser.selectedTab = gBrowser.addTab("http://privacychoice.uservoice.com/pages/18032-uninstallation");
			  break;
			case "item-cancel-action":
			  this.beingUninstalled = false;
			  break;
		  }
        }
        break;
      case "xpcom-shutdown":
        observerService.removeObserver(this,"xpcom-shutdown");
        observerService.removeObserver(this, "em-action-requested");
        observerService.removeObserver(this, "profile-after-change");
        observerService.removeObserver(this, "profile-before-change");
        break;

    }
    return;
  },
  classDescription: "OptOut Service",
  contractID: "@privacychoice.org/optout-service;1",
  classID: Components.ID('{75c6c32f-10bf-48d6-945a-57f0c610ccc0}'),
  QueryInterface: XPCOMUtils.generateQI([Ci.IOptOut, Ci.nsIObserver]),
  _xpcom_categories: [{
    category: "app-startup",
    service: true
  }]
}

function NSGetModule(compMgr, fileSpec) {
  return XPCOMUtils.generateModule([OptOut]);
}
